/** APSystems - read PV-installation from ECU-R (and maybe other ECU types too)
  * Ewald Kendziorra sr, www.kendziorra.nl, 2022
  * totally free to use and modify, no support
  * suported inverters: DS3, YC600, QS1, YC1000 (short: DS3,YC6,QS1,YC1)
  * APS ECU commands are extracted and analyzed from the "APsystems_EMA_App_v8.0.1.apk"
  *   And INFO FROM - Local Network Interface commands: 
  *   APsystems-Energy-Communication-Unit-ECU-C-User-manual_Rev1.5_2018-1-16.pdf                    **/
/** ############################################################################################### *
  * ##### For storage and hystory you better use my P1 Dutch Smart Meter (P1 DSRM) project    ##### *
  * ##### That is more accurate and you have all your energy data from previous and this year ##### *
  * ##### together, including graphs and tables. http://www.kendziorra.nl/arduino-and-esp/123 ##### *
  * ############################################################################################### *
  * ##### FIRST TO DO: import my "simpleLib" library http://www.kendziorra.nl                 ##### *
  * ############################################################################################### **/

/******************************************
  * use these directives for testing only *
  *****************************************/
//#define USE_TEST      // starts a one-time test of all commands in setup(), then the program stops ...
                        // ... this ignore commands from a Websocket Client ...
                        // ... select USE_MONITOR1 for output to Serial Monitor with description
//#define USE_MONITOR1  // Websocket output also to Serial Monitor
//#define USE_MONITOR2  // show ECU write and read buffers in Serial Monitor
/******************************************/
///// simpleLib directives
#define USE_STA
#define USE_STAIP
#define USE_STADNS
#define USE_OTA   // optional, not using? comment it
#define USE_TIME
#define USE_WS    // optional, not using? comment it
//#define USE_JSON  // optional, not using? comment it  -  change and add your own code

// STRING INFORMATION:
//  An ESP32 has so much storage space that we can quietly use String instead of *char or char[] ;=)
//  And so it is not necessary to use the flash based F(<sting>) option.
//  send buffer: We use a HEX-string to send bytes.
//  read buffer: The easiest way to decode a byte-stream is to put it in a HEX-string.
//    HEX-string: each byte as "00"->"FF" - e.g. uint8_t {1, 127, 12, 253} as String "017F0CFD"
//  Output in JSON format? look at doJson.ino example, doJson(<data>) this split and format doSyst() and doNow() data

/////////////////////////////
///// START EDIT REGION /////  
/////////////////////////////
///// Your APS ECU data: ECU_n module ID, Micro-Inverters ID, httpPORT and WiFi/LAN IP
const String
  ecuId = "2162000nnnnn"@,  // the top/first BAR-code on your ECU (back side)
  // we know which inverters we have so we don't have to look for them in the ECU
  // Inverter order: The order as you want to read them  (DC3[2], YC600[2], QS1[4], YC1000[3] - [n]: max. number of PV-panels)
  inv[] = {"4080001nnnnn"@, "8060000nnnnn"@, "4080001nnnnn"@", "8060000nnnnn"@};  // Inverters ID from the labels, maximal 18 PV-panels (2, 3 or 4 PV-panels/Inverter)
  @//      garage-YC600-2   garage-QS1-4   schuur-YC600-2   schuur-QS1-4        (placement to remember)
const uint16_t httpPort = 8899;  // the standard ECU IP-port
const char *host = "192.168.178.33"@;  // insert your local ECU IP (search it in your router and make it STATIC) ...
                                      // ... use WiFi-IP or LAN-IP  (both should be able to do the job)
/////////////////////////////
///// simpleLib data
const char                                          
      *staSsid = "edit-edit",@     // your SSID ...
  *staPassword = "edit-edit",@;  // ... and password
const IPAddress
         staIP(192, 168, 178, 128)@,  // Set your own IP also in the HTML-files and as static in your router !!!!!!!!!!!!
  staGatewayIP(192, 168, 178,   1)@,  //   ''      ''
   staSubnetIP(255, 255, 255,   0),  // use this unchanged
    staPrimDNS(8, 8, 8, 8),  // used with NTP-server(s) (time servers)...
     staSecDNS(8, 8, 4, 4);  // ... likely to be used unchanged
#ifdef USE_OTA
  const char *myOtaID = "ECU_ESP32";  // OTA name (Over-The-Air firmware upload name)
#endif
#ifdef USE_WS
  const char                         
        *wsURI = "/",          // WebSocket URI,       also in HTML
    *wsProtocol = "ECU_ESP32";  // WebSocket Protocol,  also in HTML
  const uint8_t wsPort = 82;    // Websocket Port,      also in HTML
#endif
const double                       
  latSun = 52.1234,@  // your latitude and longitude location (GPS) ...
  lonSun =  5.5678;@  // ... to get the sun day-rise/set times in seconds
/** time ***** List of timezones https://sites.google.com/a/usapiens.com/opnode/time-zones **/
char timeZone[] = @"CET-1CEST-2,M3.5.0/02:00:00,M10.5.0/03:00:00";  // EU Amsterdam/Barcelona with DST/Summertime
// char timeZone[] = "CET-1";       // EU Amsterdam/Barcelona - NO DST/Summertime 
const time_t syncTime = 12 * 3600;  // NTP time sync interval in sec. if 0 then NOT used
///////////////////////////
///// END EDIT REGION /////
///////////////////////////
// after the simpleLib declarations!!!
#include <sWiFiFull.h>  // this and the definitions above take a lot of work off your hands 

const uint8_t invNr = sizeof(inv) / sizeof(inv[0]);  // calc the number of INVERTERS
String
  setBuf = "",  // write buffer to   ECU
  getBuf = "",  // read  buffer from ECU
     ymd = "",  // "yyyymmdd" used by doDay()
  strSign[invNr] = {""};  // Zigbee signal strength, used by doNow()
time_t t;  // local time (UNIX timestamp) from my simpleLib

WiFiClient client;
sRiseSet rs;  // sun rise/set object, time in seconds of the day - from my simpleLib

///////////////////////
// format/write/read
void doSetDat(String cmd);  // format send buffer
void doWriteRead();         // command to ECU and answer from ECU...r)
// read data from ECU
void doSyst();  // kWh: up to now,  W: at this time, kWh: today so far
void doNow();   // inverters and PV-panels values
void doSign();  // Zigbee signal levels
void doDay();   // a 5 minutes list of this day;yyyymmdd Watt values - max +/- 70 days back or now
void doHyst();  // days power:  the last +/- 28 days, incl. the today fraction
// helpers
void wait(time_t w);  // dealay() alternative
bool doTest();        // data in read buffer OK?
String getTime(uint16_t start);              // get a hh:mm string      from HEX read buffer
String getStr(uint16_t start, uint8_t len);  // get a string               ...      ...
double getVal(uint16_t start, uint8_t len);  // get a value                ...      ...
String addStr(double val, int8_t dec);       // get a string from double   ...      ...
// information in/out
void info(String str);  // send data to Websocked (and Serial?)
#ifdef USE_WS
  void wsAction();  // message from a WebSocket received? - from my simpleLib
#endif
#ifdef USE_JSON
  String doJson(String _buf);  // format doSyst() or doNow() result to a JSON string
#endif
///////////////////////

void setup() {
  #ifdef USE_MONITOR1
    Serial.begin(115200);  // only for testing
    wait(2000);
    Serial.println("\n");
  #endif
  setupWiFi();  // simpleLib: init all the definitions
  t = now();
  #ifdef USE_TEST  // output to websocket and, if USE_MONITOR1 is active, also to monitor
    doSyst();
    doNow();
    time_t dt = t - 24 * 3600;  // yesterday
    char c[10];
    sprintf(c, "%04u%02u%02u", year(dt), month(dt), day(dt));
    ymd = c;
    doDay();
    doHyst();
    while(1){;}  // endless nothing
  #endif
}

void loop() {
  static uint8_t saveMinute = 99;  // new minute? save on SD - new day/read ECU
  static bool
       setRs = true, // avoid one minute long reps (sun rise/set times)
    darkness = true;  // one time it's darkness flag
  loopWiFi();  // simpleLib: test and/or carry out assignments (always the first action in loop())
  #ifdef USE_WS  
    if (wsVal) wsAction(); // read WebSocket (wsVal: a notification from my simpleLib library)
  #endif
  t = now();
  if ((hour(t) == 0 && minute(t) == 0) || setRs == true) {
    rs = sSun(t, latSun, lonSun);  // get day time sun rise/set
    setRs = false;  // avoid one minute long reps
  }
  if (minute(t) != 0) setRs = true;  // new sun rise/set test at 00:00:00
  if (saveMinute != minute(t) && ((minute(t) % 5) == 0)) { // read every 5 minutes and first time
    saveMinute = minute(t);    
    if (sDaylight(t, rs)) {  // only on daylight
      if (wsClients > 0) doSyst();
      darkness = false;
    }
    else if (!darkness) {
      darkness = true;
      info("---darkness---");
    }
  }
}

void wait(time_t w) {  // dealay() alternative
  vTaskDelay(pdMS_TO_TICKS(w));  // a NOT blocked delay()
}

void doSetDat(String cmd) {  // format send buffer
  /**
    x: String ecuId   
    z: "END"
    r: "\n"
    y: "yyyymmdd" from String ymd
  **/      
  setBuf = "APS" + cmd;
  setBuf.replace("z", "END");
  setBuf.replace("r", "\n");
  if (setBuf.indexOf("x")) setBuf.replace("x", ecuId);
  if (setBuf.indexOf("y")) setBuf.replace("y", ymd);  // ymd yyyymmdd set by user
  doWriteRead();
  if (getBuf == "") {  // no data? try again?
    wait(1000);  // wait a while and then again
    doWriteRead();
  }
  #ifdef USE_MONITOR2
    Serial.print("\nsetBuf: " + setBuf + "getBuf: " + getBuf);
    // info("\nsetBuf: " + setBuf + "getBuf: " + getBuf);        
  #endif
}

void doWriteRead() {  // command to ECU and answer from ECU...r)
  char c[3];
  time_t timeout;
  getBuf = "";
  if (client.connect(host, httpPort)) {
    wait(100);
    client.print(setBuf);
    timeout = millis() + 5000;
    while (timeout > millis()) {
      while (client.available()) {
        byte b = client.read();
        sprintf(c, "%02X", b);
        getBuf += c;
        timeout = millis() + 500;
      }
    }
    client.stop();
  }
}

void doSyst() {  // kWh: up to now,  W: at this time, kWh: today so far
  doSetDat("1100160001zr");
  bool f = doTest();
  String strSyst = "SYST;" + getBuf.substring(78, 86) + ";" + getTime(86);  // yyyymmdd;hh:mm
  if (f) {
    strSyst += addStr(getVal(54, 8) / 10.0,  2);  // kWh - up to now
    strSyst += addStr(getVal(62, 8), 0);          //  W  - at this time
    strSyst += addStr(getVal(70, 8) / 100.0, 2);  // kWh - today so far
  }
  else strSyst += "no data";
  #ifdef USE_MONITOR1
    info("\ncommand: Syst (System information)\n     date     time kWh (LifeTime), W (now), kWh (today)");
  #endif 
  info(strSyst);  // or your own way (split and/or JSON format? look at test.ino)
  #ifdef USE_JSON
    String str = doJson(strSyst);  // format doSyst() result to a JSON string  
    info(str);  // do your own thing with this 'JSON str' - now as example to Serial/Websocket monitor
  #endif    
}

void doNow() {  // inverters and PV-panels values
  doSign();  // Signal level firts
  doSetDat("1100280002xzr");
  String strNow[invNr];
  bool f = doTest();
  if (f) {
    int16_t pos;
    uint8_t panels;
    uint16_t voltage[2];
    String ts = getBuf.substring(38, 46) + ";" + getTime(46) + ";";
    for (uint8_t i = 0; i < invNr; i++) {
      String invType = "";
      String invName = "";
      String s = "";
      strNow[i] = "NOW" + String(i) + ";" + ts + inv[i];  // date, time and set inverter ID
      pos = getBuf.indexOf(inv[i]);  // sears for inverter[n]
      if (pos == -1) {
        strNow[i] += ";Inverter not found";
        continue;  // skip this inverter
      }
      if (getVal(pos += 12, 2) != 1) {  // this inverter online?
        strNow[i] += ";Darknes or offline";
        continue;  // skip this inverter
      }
      strNow[i] += strSign[i];  // from doSign()
      invType = getStr(pos += 2, 4);  // inverter type and number of panels
      if (invType == "01") {
        invName = (inv[i].startsWith("4") ? ";YC6" : ";DC3");
        panels = 2;
      }
      else if (invType == "02") {
        invName = ";YC1";
        panels = 3;
      }
      else if (invType == "03") {
        invName = ";QS1";
        panels = 4;
      }
      else {
        s += ";???";
        continue;  // skip this inverter
      }
      s += addStr(getVal(pos += 4, 4) / 10.0, 0);  // Hz
      s += addStr(getVal(pos += 4, 4) - 100.0, 0);  // *C
      voltage[0] = voltage[1] = 0;
      if (panels <= 3) {
        for (uint8_t j = 0; j < panels; j++) {  // 2 or 3 panels
          s += addStr(getVal(pos += 4, 4), 0);  // W
          voltage[0] += (uint16_t)getVal(pos += 4, 4); // ACV
          voltage[1]++;
        }
      }
      else {
        for (uint8_t j = 0; j < panels; j++) {  // 4 panels
          s += addStr(getVal(pos += 4, 4), 0);  // W
          if (j == 0) {
            voltage[0] = (uint16_t)getVal(pos += 4, 4);  // ACV
            voltage[1] = 1;
          }
        }
      }
      strNow[i] = strNow[i] + invName + ";" + String(voltage[0] / voltage[1]) + s; 
    }
  }
  for(uint8_t i = 0; i < invNr; i++) {
    if (!f) strNow[i] += "no data";
  }
  #ifdef USE_MONITOR1
    info("\ncommand: Now ('Zigbee signal level' and 'data of this moment')\n     date     time  inverter     db type ACV Hz *C W(1,2,3,4)");
  #endif 
  for (uint8_t i = 0; i < invNr; i++) { 
    info(strNow[i]);  // or your own way (split and/or JSON format? look at test.ino)
    #ifdef USE_JSON
      String str = doJson(strNow[i]);  // format doNow() result to a JSON string  
      info(str);  // do your own thing with this 'JSON str' - now as example to Serial/Websocket monitor
    #endif    
  }
}

void doSign() {  // Zigbee signal levels
  uint16_t j;
  doSetDat("1100280030xzr");
  if (doTest()) {
    for (uint8_t i = 0; i < invNr; i++) {
      j = getBuf.indexOf(inv[i]);  // sears for inverter[n]
      if (j == -1) strSign[i] = ";???";
      else strSign[i] = addStr(-getVal(j += 12, 2) / 2.55, 0);  // signal level 
    }
  }
  else for (uint8_t i = 0; i < invNr; i++) strSign[i]+= ";???";
  #ifdef USE_MONITOR2
    info("\ncommand: Sign ('Zigbee signal level', used in command Now)");
  #endif   
}

void doDay() {  // a 5 minutes list of this day-yyyymmdd Watt values - max +/- 70 days back or now
  doSetDat("1100390003xzyzr");
  String strDay = ymd;  // yyyymmdd
  if (ymd.length() == 8) {
    if (doTest()) {
      getBuf = getBuf.substring(30, getBuf.indexOf("454E440A"));  // strip start/end data
      if ((getBuf.length() % 8) != 0) strDay += ";Data dirty";
      else if (getBuf.length() < 8)  strDay += ";Wrong date";
      else {
        for (uint16_t i = 0; i < getBuf.length(); i += 8) {
          strDay = strDay + ";" + getBuf.substring(i, i + 2) + ":" + getBuf.substring(i + 2, i + 4);  // hh:mm
          strDay += addStr(getVal(i + 4, 4), 0);  //  W  - at this time
        }
      }
    }
    else strDay += ";no data";
  }
  else strDay += ";Wrong date";  
  #ifdef USE_MONITOR1
    info("\ncommand: Day or Day;yyyymmdd (Day power, 5 minute interval, up to +/- 70 days ago)\n    yyyymmdd time  W ...");
  #endif 
  info("DAY;" + strDay);  // or your own way (split and/or JSON format? look at test.ino)
}

void doHyst() {  // days kWh power: the last +/- 28 days, incl. the today fraction
  // actually the summed data as from doDay()
  doSetDat("1100330004xz01zr");
  String
    strHyst = "",
          c = "";
  if (doTest()) {
    getBuf = getBuf.substring(50, getBuf.indexOf("454E440A"));  // strip [start][first-date][end] data
    if ((getBuf.length() % 16) != 0) strHyst += ";Data dirty";
    else if (getBuf.length() < 15 * 16)  strHyst += ";Wrong date";
    else {
      for (uint16_t i = 0; i < getBuf.length(); i += 16) {
        strHyst = strHyst + c + getBuf.substring(i, i + 8);  // yyyymmdd
        strHyst += addStr(getVal(i + 8, 8) / 100.0, 2);  // kWh  - at this date
        c = ";";
      }
    }
  }
  else strHyst += ";no data";  
  #ifdef USE_MONITOR1
    info("\ncommand: Hyst (the last +/- 28 days, incl. the today fraction)\n     yyyymmdd kWh ...");
  #endif 
  info("HYST;" + strHyst);  // or your own way (split and/or JSON format? look at test.ino)
}

/** helpers **/

bool doTest() {  // data in read buffer OK?
  bool f = getBuf.length() == (getStr(10, 8).toInt() * 2 + 2);
  if (f) f = getBuf.startsWith("415053") && getBuf.endsWith("454E440A");
  return f;
}

String getTime(uint16_t start) {  // get a hh:mm string from HEX read buffer
  return getBuf.substring(start, start + 2) + ":" + getBuf.substring(start + 2, start + 4);  
}

String getStr(uint16_t start, uint8_t len) {  // get a string from HEX read buffer
  char c;
  String s = "";
  for (uint16_t i = start; i < start + len; i += 2)  {
    c = strtoul(getBuf.substring(i, i + 2).c_str(), NULL, 16);
    s += c;
  }
  return s;
}

double getVal(uint16_t start, uint8_t len) {  // get a value HEX read buffer
  return String("0x" + getBuf.substring(start, start + len)).toDouble();
}

String addStr(double val, int8_t dec) {  // get a string from double HEX read buffer
  return ";" + String(val, dec);
}

void info(String str) {  // send data to Websocked (and Serial?)
  #ifdef USE_MONITOR1
    Serial.println(str);
  #endif
  #ifdef USE_WS  
    webSocket.broadcastTXT(str + "\n");  // or your own way
  #endif
}

#ifdef USE_WS
  void wsAction() {  // message from a WebSocket received? - from my simpleLib
    wsVal = false; // reset notification
    // wsBuf: variable from simpleLib.h Websocket section
    // Websocket command from user or ...: Syst, Now, Day, Hyst
    if (wsBuf == "Syst") doSyst();  // send doSyst() (also automatic send in 5 minutes interval)
    if (wsBuf == "Now") doNow();  // send now doNow()
    if (wsBuf == "Day") { // send this day date doDay()
      char c[10];
      sprintf(c, "%04u%02u%02u", year(t), month(t), day(t));
      ymd = c;
      doDay();
    }
    if (wsBuf.startsWith("Day;")) {  // send yyyymmdd date doDay()
      ymd = wsBuf.substring(4);  // yyyymmdd format
      doDay();
    }
    if (wsBuf == "Hyst") doHyst();  // send doHyst()
  }
#endif

#ifdef USE_JSON
  String doJson(String _buf) {  // format doSyst() or doNow() result to a JSON string
    const String
      _syst[5] = {"date", "time", "live", "now", "day"},
      _now[12] = {"date", "time", "inv", "db", "type", "AVC", "Hz", "*C", "p1W", "p2W", "p3W", "p4W"};
    String
      js = "",
      split;
    uint8_t i, n = 0, pos[13];
    int8_t p = _buf.indexOf(";");
    while (p != -1) {
      pos[n++] = ++p;
      p = _buf.indexOf(";", p);
    }
    if (n == 5 || (n >= 10 && n <= 12)) {
      pos[n] = _buf.length() + 1;
      js = "{";
      for (i = 0; i < n; i++) {
        split = _buf.substring(pos[i], pos[i + 1] - 1);
        js = js + "\"" + (n == 5 ? _syst[i] : _now[i]) + "\":\"" + split + "\"" + (i < (n - 1) ? "," : "");
      }
      js += "}";
    }
    return js;  // js.length == 0: ERROR
  }
#endif